# check_config.py -- Validate P4DTI configuration parameters.
#
# Nick Barnes, Ravenbrook Limited, 2001-01-18
#
# $Id: //info.ravenbrook.com/project/p4dti/version/1.0/code/replicator/check_config.py#4 $
# 
# This file is copyright (c) 2001 Perforce Software, Inc.  All rights  
# reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1.  Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
# 
# 2.  Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

import re
import types

error = "P4DTI configuration error"

def check_bool(param, name):    
    if (param != 0 and param != 1):
        raise error, ("Configuration parameter '%s' must be 0 or 1." %
                      name)

def check_date(param, name):
    check_string(param, name)
    date_re = "^\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d$"
    if not re.match(date_re, param):
        raise error, ("Configuration parameter '%s' (value '%s') is not a "
                      "valid date.  The right format is 'YYYY-MM-DD HH:MM:SS'."
                      % (name, param))        

def check_email(param, name):
    check_string(param, name)

    # RFC 822 defines an 'addr-spec' as follows.
    #
    #   addr-spec   =  local-part "@" domain
    #   local-part  =  word *("." word)
    #   word        =  atom
    #   domain      =  sub-domain *("." sub-domain)
    #   sub-domain  =  domain-ref
    #   domain-ref  =  atom
    #   atom        =  1*<any CHAR except specials, SPACE and CTLs>
    #   specials    =  "(" / ")" / "<" / ">" / "@" /  "," / ";" / ":" / "\"
    #   		   / <"> /  "." / "[" / "]"
    #
    # (A CHAR is an ASCII character.)
    #
    # This isn't the full story.  "Structured field bodies" in RFC822 also
    # permit quoting, domain literals and comments.  This code doesn't
    # recognise these features.  So users of the P4DTI will have to get by
    # without.
    atom_re = "[!#$%&'*+\\-/0-9=?A-Z^_`a-z{|}~]+"
    email_re = "^%s(\\.%s)*@%s(\\.%s)*$" % (atom_re, atom_re, atom_re, atom_re)

    if not re.match(email_re, param):
        raise error, ("Configuration parameter '%s' (value '%s') is not a "
                      "valid e-mail address." % (name, param))

def check_function(param, name):
    if not isinstance(param, types.FunctionType):
        raise error, ("Configuration parameter '%s' must be a function." %
                      name)

def check_host(param, name):
    check_string(param, name) # could do more, e.g. look it up.

def check_int(param, name):    
    if not isinstance(param, types.IntType):
        raise error, ("Configuration parameter '%s' must be an integer." %
                      name)

def check_list_of(param, name, type, typename):
    if not isinstance(param, types.ListType):
        raise error, ("Configuration parameter '%s' must be a list." % name)
    for item in param:
        if not isinstance(item, type):
            raise error ("Configuration parameter '%s' must be a list of %s." %
                         (name, typename))

def check_string(param, name):
    if not isinstance(param, types.StringType):
        raise error, ("Configuration parameter '%s' must be a string." %
                      name)

def check_string_or_none(param, name):
    if not (param == None or
            isinstance(param, types.StringType)):
        raise error, ("Configuration parameter '%s' must be None or a string."
                      % name)

# A. REFERENCES
#
# [RFC 822] "RFC 822: Standard for the format of ARPA Internet text messages";
# David H Crocker; 1982-08-13; <ftp://src.doc.ic.ac.uk/rfc/rfc822.txt>.
#
#
# B. DOCUMENT HISTORY
#
# 2001-01-18 NB Moved from configure_bugzilla.py so we can share with other
# DTs.
#
# 2001-01-23 GDR Extended check_email so that it checks RFC822 address syntax.
#
# 2001-02-04 GDR Alphabetized.  Added check_date.
#
# 2001-02-16 NB Added check_function (for checking replicate_p parameter).
#
# 2001-03-02  RB    Transferred copyright to Perforce under their license.
