# configure_bugzilla.py -- Build P4DTI configuration for replication
# to Bugzilla.
#
# Richard Brooksby, Ravenbrook Limited, 2000-12-07
#
# $Id: //info.ravenbrook.com/project/p4dti/version/1.0/code/replicator/configure_bugzilla.py#15 $
#
#
# 1. INTRODUCTION
#
# This is the automatic configurator for the Bugzilla integration of the
# Perforce Defect Tracking Integration (P4DTI).
#
# This code takes a basic set of configuration parameters [RB 2000-08-10, 5.1]
# and generates a full set of parameters for the "dt_bugzilla", "p4", and
# "replicator" classes, as well as a Perforce jobspec.
#
# ---
#
# This file is copyright (c) 2001 Perforce Software, Inc.  All rights  
# reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1.  Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
# 
# 2.  Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

import dt_bugzilla
import logger
import replicator
import socket
import string
import bugzilla
import MySQLdb
import check_config
import types
import os
import keyword

error = "Bugzilla configuration error"

# Force MySQLdb to use plain strings as date/time types
try:
    from DateTime import DateTimeType, DateTimeDeltaType
    # types of values coming out of MySQL
    for t in [MySQLdb.FIELD_TYPE.DATETIME,
              MySQLdb.FIELD_TYPE.DATE,
              MySQLdb.FIELD_TYPE.TIME,
              MySQLdb.FIELD_TYPE.TIMESTAMP]:
        if MySQLdb.type_conv.has_key(t):
            del MySQLdb.type_conv[t]
    # types of values going into MySQL
    for t in [DateTimeType,
              DateTimeDeltaType]:
        if MySQLdb.quote_conv.has_key(t):
            del MySQLdb.quote_conv[t]
except ImportError:
    # e.g. no DateTime module; MySQLdb will revert to plain strings anyway.
    pass

# perforce keyword translator.
keyword_translator = keyword.keyword_translator()

# enum translator (just like keyword translator except '' <-> 'NONE')
enum_translator = dt_bugzilla.enum_translator(keyword_translator)

# make_state_pairs: (strings * string) -> (string * string) list.  Make a
# list of pairs of state names (Bugzilla state, Perforce state).  This list
# will be used to translate between states, and also to generate the possible
# values for the Status field in Perforce.
#
# The closed_state argument is the Bugzilla state which maps to the special
# state 'closed' in Perforce, or None if there is no such state.  See
# requirement 45.  See decision decision [RB 2000-11-28 14:44:36 GMT].
#
# The case of state names in these pairs is normalized for usability in
# Perforce: see design decision [RB 2000-11-28 14:24:32 GMT].

def make_state_pairs(states, closed_state):
    state_pairs = []
    state_p4_to_dt = {}
    found_closed_state = 0
    if closed_state != None:
        p4_closed_state = keyword_translator.translate_0_to_1(
            string.lower(closed_state))

    # Perforce jobs can't have state "new" (this indicates a fresh job and
    # Perforce changes the state to "open").  Nor can they have state
    # "ignore", because that is used in the submit form to indicate that a
    # job shouldn't be fixed by the change.
    # 
    # Unfortunately, "new" and "ignore" are common names for states in
    # the defect tracker (the former is in the default workflow in
    # TeamTrack), so we don't disallow them, but prefix them with
    # 'bugzilla_'.  Then we quit if two Bugzilla states map to the
    # same state in Perforce, ruling out the unlikely situation that
    # someone has a Bugzilla status of 'BUGZILLA_CLOSED'.  See job000141.

    prohibited_states = ['new', 'ignore']
    prohibited_state_prefix = 'bugzilla_'
        
    for dt_state in states:
        p4_state = keyword_translator.translate_0_to_1(string.lower(dt_state))

        if closed_state != None:
            if p4_state == p4_closed_state:
                p4_state = 'closed'
                found_closed_state = 1
            elif p4_state == 'closed':
                p4_state = prohibited_state_prefix + p4_state

        if p4_state in prohibited_states:
            p4_state = prohibited_state_prefix + p4_state
        if (state_p4_to_dt.has_key(p4_state)
            and state_p4_to_dt[p4_state] != dt_state):
            raise error, ("Two Bugzilla states '%s' and '%s' map to the "
                          "same Perforce state '%s'."
                          % (dt_state, state_p4_to_dt[p4_state], p4_state))
        state_p4_to_dt[p4_state] = dt_state
        pair = (dt_state, p4_state)
        if pair not in state_pairs:
            state_pairs.append(pair)

    if closed_state != None and not found_closed_state:
        raise error, ("You specified the closed_state '%s', but there's no "
                      "such Bugzilla state." % closed_state)
    return state_pairs


# Given the name of an enum column, calculate the values and default.
def translate_enum(bugs_types, column):
    if not bugs_types[column]['type'] == 'enum':
        raise error, ("The '%s' column of Bugzilla's 'bugs' table "
                      "is not an enum type." % column)
    values = bugs_types[column]['values']
    values = map(enum_translator.translate_0_to_1,
                 bugs_types[column]['values'])
    default = bugs_types[column]['default']
    if default != None:
        default = enum_translator.translate_0_to_1(default)
    values = string.join(values,'/')
    return (values, default)

# Some Bugzilla fields should not be changed from Perforce, either
# because they can not be changed from Bugzilla
# (e.g. 'creation_ts', 'delta_ts', 'lastdiffed'), or because they
# can only be changed in Bugzilla in very restricted ways
# (e.g. 'groupset', 'product', 'version', 'component',
# 'target_milestone', 'everconfirmed'), or because changing them
# in Bugzilla has complex side-effects which can't be sensibly
# reproduced here (e.g. 'votes', 'keywords').

read_only_fields = ['bug_id',
                    'groupset',
                    'creation_ts',
                    'delta_ts',
                    'product',
                    'version',
                    'component',
                    'target_milestone',
                    'votes',
                    'keywords',
                    'lastdiffed',
                    'everconfirmed']

# Some Bugzilla fields can only be appended to.
# In particular, the 'Descriptions' section of a Bugzilla bug
# is actually a number of rows from another table, to which
# the Bugzilla web interface allows one to add a row.
# We fake this as if it's a regular field of the table, but
# retain the append-only restriction.

append_only_fields = ['longdesc']

# Changes to some bugzilla fields do not get recorded in the
# bugs_activity table.

fields_not_in_bugs_activity = ['longdesc']

# Map bugzilla field name to a name we can use in Perforce and a
# comment.

bz_field_map = {
    'longdesc':          ('Description',
                          "Description and comments."),
    'assigned_to':       ('Assigned_To',
                          "User to which the bug is assigned."),
    'groupset':          ('Groupset', None),
    'bug_file_loc':      ('URL',
                          "The bug's URL."),
    'bug_severity':      ('Severity',
                          "The bug's severity."),
    'bug_status':        ('Status',
                          "The bug's status."),
    'creation_ts':       ('Creation_Timestamp',
                          "Time created."),
    'delta_ts':          ('Update_Timestamp',
                          "Time last updated."), 
    'short_desc':        ('Summary',
                          "The bug's summary."),
    'op_sys':            ('OS',
                          "The OS to which the bug applies."),
    'priority':          ('Priority',
                          "The bug's priority."),
    'product':           ('Product',
                          "The product."),
    'rep_platform':      ('Platform',
                          "The hardware platform to which the bug applies."),
    'reporter':	         ('Reporter',
                          "The Bugzilla user who reported the bug."),
    'version':	         ('Version',
                          "The product version."),
    'component':	 ('Component',
                          "The product component."),
    'resolution':	 ('Resolution',
                          "The manner in which the bug was resolved."),
    'target_milestone':  ('TargetMilestone',
                          "The bug's target milestone."),
    'qa_contact':	 ('QA_Contact',
                          "The Bugzilla user who is the QA contact for this bug."),
    'status_whiteboard': ('StatusWhiteboard',
                          "The bug's status whiteboard."),
    'votes':	         ('Votes',
                          "The number of votes for this bug."),
    'keywords':	         ('Keywords',
                          "Keywords for this bug."),
    'lastdiffed':	 ('LastDiffed',
                          "Time last compared for automated e-mail."),
    'everconfirmed':     ('EverConfirmed',
                          "Has this bug ever been confirmed?"),
    'bug_id':            ('Bug_number',
                          "Bug number."),
    }

def configuration(administrator_address = None,
                  changelist_url = None,
                  closed_state = None,
                  log_file = 'p4dti.log',
                  p4_client_executable = 'p4',
                  p4_password = None,
                  p4_port = None,
                  p4_server_description = 'P4DTI Perforce server',
                  p4_user = None,
                  poll_period = None,
                  replicate_p = None,
                  replicated_fields = [],
                  replicator_address = None,
                  rid = 'p4dti',
                  sid = 'p4dti',
                  smtp_server = None,
                  start_date = None,
                  dbms_host = None,
                  dbms_database = None,
                  dbms_port = None,
                  dbms_user = None,
                  dbms_password = '',
                  bugzilla_directory = None,
                  verbose = 0,
                  ):

    # Validate configuration items.
    check_config.check_string(rid, 'rid')
    check_config.check_string(sid, 'sid')
    if administrator_address != None:
        check_config.check_email(administrator_address, 'administrator_address')
    check_config.check_string_or_none(log_file, 'log_file')
    check_config.check_string(p4_client_executable, 'p4_client_executable')
    check_config.check_string(p4_port, 'p4_port')
    check_config.check_string(p4_user, 'p4_user')
    check_config.check_string(p4_password, 'p4_password')
    check_config.check_string(p4_server_description, 'p4_server_description')
    check_config.check_int(poll_period, 'poll_period')
    check_config.check_function(replicate_p, 'replicate_p')
    check_config.check_email(replicator_address, 'replicator_address')
    if smtp_server != None:
        check_config.check_host(smtp_server, 'smtp_server')
    check_config.check_date(start_date, 'start_date')    
    check_config.check_bool(verbose, 'verbose')
    check_config.check_string_or_none(closed_state, 'closed_state')
    check_config.check_list_of(replicated_fields, 'replicated_fields',
                               types.StringType, 'strings')
    check_config.check_host(dbms_host, 'dbms_host')
    check_config.check_int(dbms_port, 'dbms_port')
    check_config.check_string(dbms_database, 'dbms_database')
    check_config.check_string(dbms_user, 'dbms_user')
    check_config.check_string(dbms_password, 'dbms_password')
    check_config.check_string_or_none(changelist_url, 'changelist_url')
    check_config.check_string_or_none(bugzilla_directory, 'bugzilla_directory')


    if bugzilla_directory != None:
        # strip any trailing / character
        if len(bugzilla_directory) > 1 and bugzilla_directory[-1:] == '/':
            bugzilla_directory = bugzilla_directory[:-1]
        if not os.path.isdir(bugzilla_directory):
            raise error, ("Configuration parameter 'bugzilla_directory' does "
                          "not name a directory.")
        processmail = bugzilla_directory + '/processmail'
        if not os.access(processmail, os.X_OK):
            raise error, ("Configuration parameter 'bugzilla_directory' "
                          "does not name a directory containing a processmail "
                          "script.")

    # Handle logger.  We need a list of logger objects:

    loggers = [logger.file_logger(), # to stdout
               logger.sys_logger()]  # to syslog

    if log_file != None:
        loggers.append(logger.file_logger(open(log_file, "a"))) # to log_file

    # now a single logger object which logs to all of the logger objects in
    # our list:
    logger_object = logger.multi_logger(loggers)

    # Open a connection to the Bugzilla database.  This makes a DB-API
    # v2.0 connection object.  To work with a database other than
    # MySQL, change this to make an appropriate connection object.
    # Note that in that case changes are also needed in bugzilla.py
    # where we deal with MySQL-specific types such as tinyint.
    db = MySQLdb.connect(host = dbms_host,
                         port = dbms_port,
                         db = dbms_database,
                         user = dbms_user,
                         passwd = dbms_password)

    # Make a table of configuration items to pass to Bugzilla.
    bugzilla_config_table = {}
    bugzilla_config_table['replicator_user'] = replicator_address
    bugzilla_config_table['p4_server_description'] = p4_server_description
    if changelist_url != None:
        bugzilla_config_table['changelist_url'] = changelist_url
    

    # Make a Bugzilla DB object.  Note that this same object is used
    # subsequently by the replicator itself.
    connection = bugzilla.bugzilla(db, rid, sid, logger_object,
                                   bugzilla_config_table, bugzilla_directory,
                                   verbose)

    # Get the types of the 'bugs' table from Bugzilla
    bugs_types = connection.get_types('bugs')

    # Get the types of the 'profiles' table from Bugzilla.  In particular we
    # need to know the size of the 'login_name' column.
    profiles_types = connection.get_types('profiles')
    if not profiles_types.has_key('login_name'):
        raise error, ("Bugzilla's table 'profiles' does not have a "
                      "'login_name' column")
    if profiles_types['login_name']['type'] != 'text':
        raise error, ("The 'login_name column of Bugzilla's 'profiles' table "
                      "does not have a 'text' type.")
    user_name_length = profiles_types['login_name']['length']

    required_columns = ['bug_status', 'resolution', 'assigned_to', 'short_desc']
    for column in required_columns:
        if not bugs_types.has_key(column):
            raise error, ("Bugzilla's table 'bugs' does not have a '%s' column."
                          % column)

    if not bugs_types['bug_status']['type'] == 'enum':
        raise error, ("The 'bug_status' column of Bugzilla's 'bugs' table "
                      "is not an enum type.")
    if not bugs_types['resolution']['type'] == 'enum':
        raise error, ("The 'resolution' column of Bugzilla's 'bugs' table "
                      "is not an enum type.")
    if not 'FIXED' in bugs_types['resolution']['values']:
        raise error, ("The 'resolution' column of Bugzilla's 'bugs' table "
                      "does not have a 'FIXED' value.")

    # Make a list of (Bugzilla state, Perforce state) pairs.
    state_pairs = make_state_pairs(bugs_types['bug_status']['values'],
                                   closed_state)

    # Make a list of possible resolutions.
    (resolutions,default_resolution) = translate_enum(bugs_types,
                                                      'resolution')

    user_translator = dt_bugzilla.user_translator(replicator_address,
                                                  p4_user,
                                                  allow_unknown = 0)

    # Work out the legal values of the State field in the jobspec.  Note that
    # "closed" must be a legal state because "p4 fix -c CHANGE JOBNAME" always
    # sets the State to "closed" even if "closed" is not a legal value.  See
    # job000225.
    legal_states = map((lambda x: x[1]), state_pairs)
    if 'closed' not in legal_states:
        legal_states.append('closed')
    state_values = string.join(legal_states, '/')
    
    # p4_fields maps Bugzilla field name to the jobspec data (number, name,
    # type, length, dispositon, preset, values, help text, translator).  The
    # five fields 101 to 105 are predefined because they are needed by both
    # Bugzilla and Perforce and the replicator.  The fields Job and Date are
    # special: they are not replicated from TeamTrack but are required by
    # Perforce, so we have them here.  Note that their help text is given (the
    # other help texts will be fetched from TeamTrack).
    p4_fields = { \
        '(JOB)':       ( 101, 'Job', 'word', 32, 'required',
                         None, None,
                         "The job name.",
                         None ),
        'bug_status':  ( 102, bz_field_map['bug_status'][0], 'select',
                         bugs_types['bug_status']['length'],
                         'required',
                         state_pairs[0][1],
                         state_values,
                         bz_field_map['bug_status'][1],
                         dt_bugzilla.status_translator(state_pairs) ),
        'assigned_to': ( 103, bz_field_map['assigned_to'][0], 'word',
                         user_name_length, 'required',
                         '$user', None,
                         bz_field_map['assigned_to'][1],
                         user_translator),
        '(DATE)':      ( 104, 'Date', 'date', 20, 'always',
                         '$now', None,
                         "The date this job was last modified.",
                         None ),
        'short_desc':  ( 105, bz_field_map['short_desc'][0], 'text',
                         bugs_types['short_desc']['length'], 'required',
                         '$blank', None,
                         bz_field_map['short_desc'][1],
                         dt_bugzilla.text_translator() ),
        'resolution':  ( 106, bz_field_map['resolution'][0], 'select',
                         bugs_types['resolution']['length'], 'required',
                         default_resolution,
                         resolutions,
                         bz_field_map['resolution'][1],
                         enum_translator),
        '(FILESPECS)': ( 191, 'P4DTI-filespecs', 'text', 0, 'default',
                         None, None,
                         "Associated filespecs.",
                         None ),
        '(ACTION)':    ( 192, 'P4DTI-action', 'select', 32, 'required',
                         'replicate', 'keep/discard/wait/replicate',
                         "Replicator action. Do not edit!",
                         None ),
        '(RID)':       ( 193, 'P4DTI-rid', 'word', 32, 'required',
                         'None', None,
                         "P4DTI replicator identifier. Do not edit!",
                         None ),
        '(ISSUE)':     ( 194, 'P4DTI-issue-id', 'word', 32, 'required',
                         'None', None,
                         "Bugzilla issue database identifier. Do not edit!",
                         None ),
        '(USER)':      ( 195, 'P4DTI-user', 'word', 32, 'always',
                         '$user', None,
                         "Last user to edit this job. You can't edit this!",
                         None ),
        }

    # The other replicated fields will be sequential from this field id.
    p4_field_id = 110

    # Go through the replicated_fields list, build structures and add them to
    # p4_fields.
    for dt_field_name in replicated_fields:
        # Field must not be already in the p4_fields structure.
        if p4_fields.has_key(dt_field_name):
            if dt_field_name in ['bug_status', 'assigned_to',
                                 'short_desc', 'resolution']:
                raise error, ("Field '%s' specified in 'replicated_fields' "
                              "is a system field: leave it out!"
                              % dt_field_name)
            else:
                raise error, ("Field '%s' appears twice in "
                              "'replicated_fields'." % dt_field_name)

        # Field must exist in Bugzilla
        if not bugs_types.has_key(dt_field_name):
            raise error, ("Field '%s' specified in 'replicated_fields' list "
                          "not in Bugzilla 'bugs' table." % dt_field_name)

        type = bugs_types[dt_field_name]
        dt_field_type = type['type']
        p4_field_values = None
        p4_field_length = None

        if bz_field_map.has_key(dt_field_name):
            p4_field_name = bz_field_map[dt_field_name][0]
            p4_field_comment = bz_field_map[dt_field_name][1]
        else:
            p4_field_name = "bugzilla_" + dt_field_name
            p4_field_comment = None

        if p4_field_comment == None:
            p4_field_comment = ("Bugzilla's '%s' field" % dt_field_name)

        # if there is a default, use it.
        if (type.has_key('default') and
            type['default'] != None and
            type['default'] != ''):
            p4_field_class = 'default'
            p4_field_default = type['default']
        else:
            p4_field_class = 'optional'
            p4_field_default = None

        if dt_field_type == 'other':
            raise error, ("Field '%s' specified in 'replicated_fields' list "
                          "has type '%s': "
                          "this is not yet supported by P4DTI."
                          % (dt_field_name, type['sql_type']))
        elif dt_field_type == 'float':
            raise error, ("Field '%s' specified in 'replicated_fields' list "
                          "has floating-point type: "
                          "this is not yet supported by P4DTI." % dt_field_name)
        elif dt_field_type == 'user':
            p4_field_type = 'word'
            p4_field_length = user_name_length
            translator = user_translator
        elif dt_field_type == 'enum':
            p4_field_type = 'select'
            (p4_field_values, p4_field_default) = translate_enum(bugs_types,
                                                                 dt_field_name)
            translator = replicator.translator()
        elif dt_field_type == 'int':
            p4_field_type = 'word'
            translator = dt_bugzilla.int_translator()
        elif dt_field_type == 'date':
            p4_field_type = 'date'
            p4_field_length = 20
            translator = dt_bugzilla.date_translator()
        elif dt_field_type == 'timestamp':
            p4_field_type = 'date'
            p4_field_length = 20
            translator = dt_bugzilla.timestamp_translator()
        elif dt_field_type == 'text':
            p4_field_type = 'text'
            translator = dt_bugzilla.text_translator()

        # "p4 -G" uses the field "code" to indicate whether the Perforce
        # command succeeded or failed.  See job000003.
        if p4_field_name == 'code':
            raise error, ("You can't have a field called 'code' in the "
                          "Perforce jobspec.")

        # Fixed-length fields get the length from Bugzilla.
        if p4_field_length == None:
            p4_field_length = type['length']

        if dt_field_name in read_only_fields:
            p4_field_comment = (p4_field_comment +
                                "  DO NOT MODIFY.")

        if dt_field_name in append_only_fields:
            p4_field_comment = (p4_field_comment +
                                "  ONLY MODIFY BY APPENDING.")

        p4_fields[dt_field_name] = \
            ( p4_field_id,
              p4_field_name,
              p4_field_type,
              p4_field_length,
              p4_field_class,
              p4_field_default,
              p4_field_values,
              p4_field_comment,
              translator
              )
        p4_field_id = p4_field_id + 1
        if p4_field_id >= 191:
            raise error, ("Too many fields to replicate: Perforce jobs can "
                          "contain only 99 fields.")

    # Build a jobspec, suitable for "p4 -G jobspec -i".
    jobspec = {}
    p4_fields_list = p4_fields.values()

    def p4_fields_less(x,y):
        if x[0] < y[0]:
            return -1
        elif x[0] > y[0]:
            return 1
        else:
            raise error, ("Jobspec fields '%s' and '%s' "
                          "have the same number %d" %
                          (x[1],y[1],x[0]))
    
    p4_fields_list.sort(p4_fields_less)

    # We really do want this code to do this.
    # See the result of p4.p4.run("p4 jobspec -o").
    i = 0
    for field in p4_fields_list:
        jobspec['Fields%d' % i] = "%s %s %s %s %s" % field[0:5]
        i = i + 1

    i = 0
    for field in p4_fields_list:
        if field[6] != None:
            jobspec['Values%d' % i] = "%s %s" % (field[1], field[6])
            i = i + 1

    i = 0
    for field in p4_fields_list:
        if field[5] != None:
            jobspec['Presets%d' % i] = "%s %s" % (field[1], field[5])
            i = i + 1

    def comment(field):
        return "# %s: %s\n" % (field[1], field[7])
    
    jobspec['Comments'] = ("# DO NOT CREATE NEW JOBS IN PERFORCE. "
                           "USE BUGZILLA.\n#\n" +
                           string.join(map(comment, p4_fields_list), ""))

    # Build Bugzilla and replicator configuration dictionaries.
    bugzilla_config = \
        { 'logger': logger_object,
          'rid' : rid,
          'sid' : sid,
          'bugzilla' : connection,
          'jobname-function': lambda bug: 'bug%d' % bug['bug_id'],
          'verbose': verbose,
          'append-only-fields' : append_only_fields,
          'read-only-fields' : read_only_fields,
          'fields-not-in-bugs-activity' : fields_not_in_bugs_activity,
          'start-date' : start_date,
          }

    # configured_fields is a list of triples (Bugzilla database field name,
    # Perforce field name, translator) required by the replicator.  We use the
    # filter to remove the fields that aren't replicated: these have no
    # translator.
    configured_fields = \
        map(lambda item: (item[0], item[1][1], item[1][8]),
            filter(lambda item: item[1][8] != None, p4_fields.items()))

    p4_config = \
        { 'p4-client': 'p4dti-%s' % socket.gethostname(),
          'p4-client-executable': p4_client_executable,
          'p4-password': p4_password,
          'p4-port': p4_port,
          'p4-user': p4_user,
          'logger' : logger_object,
          'verbose': verbose,
          }
    
    replicator_config = \
        { 'administrator-address': administrator_address,
          'date-translator': dt_bugzilla.date_translator(),
          'job-owner-field': 'Assigned_To',
          'job-status-field': 'Status',
          'logger': logger_object,
          'poll-period': poll_period,
          'replicate-p': replicate_p,
          'replicated-fields': configured_fields,
          'replicator-address': replicator_address,
          'smtp-server': smtp_server,
          'text-translator': dt_bugzilla.text_translator(),
          'user-translator': dt_bugzilla.user_translator(replicator_address,
                                                         p4_user,
                                                         allow_unknown = 1),
        }

    return bugzilla_config, p4_config, replicator_config, jobspec

# A. REFERENCES
#
# [RB 2000-08-10] "Perforce Defect Tracking Integration Administrator's
# Guide"; Richard Brooksby; Ravenbrook Limited; 2000-08-10.
#
#
# B. DOCUMENT HISTORY
#
# 2000-12-07 RB Branched and adapted from configure_teamtrack.py.
#
# 2000-12-08 GDR Translate state name "ignore" to "_ignore".
#
# 2000-12-13 NB State names need to be lower case in Perforce.  Also logger
# object needed for the bugzilla.bugzilla object.
#
# 2000-12-15 NB Added verbosity support.
#
# 2001-01-11 NB Added support for replicated_fields.  Also support
# closed_state, and move MySQL connection code here (out of bugzilla.py)
# so that we only open one connection when starting up.  Also sorted out
# the size of the jobspec fields.
#
# 2001-01-12 NB Moved configuration of read-only and append-only fields here
# from dt_bugzilla.py.  Added configuration of fields not recorded in the
# bugs_activity table.  Added comments for read-only and append-only fields.
#
# 2001-01-15 NB Added a table for field names and comments, because the
# automatically-generated ones were terrible.  Added validation for
# config parameters.
#
# 2001-01-18 NB Removed bugzilla_user.  Moved configuration checks out to
# check_config.py.  Pass replicator_user to Bugzilla.
#
# 2001-01-19 NB closed_state and log_file may be None.  Use system logger.
#
# 2001-01-25 NB Added bugzilla_directory to support processmail.
#
# 2001-01-26 NB Pass p4_server_description to Bugzilla.
#
# 2001-02-04 GDR Added start_date parameter.
#
# 2001-02-08 NB Prevent the existence of DateTime from changing our behaviour.
# (job000193).
#
# 2001-02-09 NB Added checks for bugzilla_directory.
#
# 2001-02-13 GDR Allow administrator_address and smtp_server to be None.
#
# 2001-02-16 NB Added replicate_p configuration parameter.
#
# 2001-02-19 NB Moved keyword translation to p4.py.
#
# 2001-02-22 GDR Moved keyword translation to keyword.py.  Made sure that
# 'closed' is a legal job state in Perforce.  Added text translator to
# replicator config.
#
# 2001-02-26 GDR Refer to "Bugzilla" explicitly in messages, not just "defect
# tracker".
#
# 2001-03-02  RB    Transferred copyright to Perforce under their license.
